/* cttest.c - parse /proc/net/ip_conntrack, compute bucket occupation
 *
 * This is how lines look:
 *
 * udp      17 0 src=213.6.73.215 dst=62.104.191.241 sport=1858 dport=53 src=62.104.191.241 dst=213.6.73.215 sport=53 dport=1858 use=1 
 * tcp      6 114 TIME_WAIT src=62.104.211.64 dst=212.7.144.66 sport=34685 dport=80 src=212.7.144.66 dst=62.104.211.64 sport=80 dport=34685 [ASSURED] use=1 
 */

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>
#include <math.h>
#include <netinet/in.h>
#include <arpa/inet.h>

typedef unsigned long long u64;
typedef unsigned long u32;
typedef unsigned short u16;
typedef unsigned char u8;

static u32 htable_size = 7168;
static u32 htable_bits = 12;

#ifdef PER_HASH_OUTER
#define PER_HASH_TIMER_1(code) do{code}while(0)
#define PER_HASH_TIMER_2(code) PER_HASH_TIMER(code)
#define PER_HASH_OUTPUT 1
#define PER_HASH_OUTPUT_TYPE "OUTER"
#else
#ifdef PER_HASH_INNER
#define PER_HASH_TIMER_1(code) PER_HASH_TIMER(code)
#define PER_HASH_TIMER_2(code) do{code}while(0)
#define PER_HASH_OUTPUT 1
#define PER_HASH_OUTPUT_TYPE "INNER"
#else
#define PER_HASH_TIMER_1(code) do{total_hash_count++;do{code}while(0);}while (0)
#define PER_HASH_TIMER_2(code) do{code}while(0)
#undef PER_HASH_OUTPUT
#endif
#endif

static u32 total_hash_count;

#ifdef PER_HASH_OUTPUT
static u64 total_hash_time;
#include "rdtsc.c"
#define PER_HASH_TIMER(code) \
	do { \
		u64 timer = rdtsc_read(); \
		do{code}while(0); \
		timer = rdtsc_read() - timer; \
		total_hash_time += timer; \
		total_hash_count++; \
	} while (0)
#endif

/*
 * The key structure is arranged to resemble the kernel tuple layout
 * as of July 2002.
 */
struct ct_key {
	u32 sip;
	u16 sport;
	u32 dip;
	u16 dport;
	u16 proto;
};

static u32 hash_seed;

/*
 * hash_conntrack(), later called "original", is the kernel code
 * as of July 2002.
 */
static u32 hash_conntrack(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ntohl(key->sip + key->dip
	+ key->sport + key->dport
	+ key->proto)
	+ ntohs(key->sport);
);
	return res;
}

/*
 * hash_rusty() is based on Rusty Russel's include/linux/hash.h
 * implementation of multiplicative hashing, as found in Linux 2.5.
 */

/* Fast hashing routine for a long.  (C) 2002 William Lee Irwin III, IBM */

#ifndef BITS_PER_LONG
#define BITS_PER_LONG 32
#endif

#if BITS_PER_LONG == 32
/* 2^31 + 2^29 - 2^25 + 2^22 - 2^19 - 2^16 + 1 */
#define GOLDEN_RATIO_PRIME 0x9e370001UL
#elif BITS_PER_LONG == 64
/*  2^63 + 2^61 - 2^57 + 2^54 - 2^51 - 2^18 + 1 */
#define GOLDEN_RATIO_PRIME 0x9e37fffffffc0001UL
#else
#error Define GOLDEN_RATIO_PRIME for your wordsize.
#endif

static inline unsigned long kernel_hash_long(unsigned long val, unsigned int bits)
{
	unsigned long hash = val;

#if BITS_PER_LONG == 64
	/*  Sigh, gcc can't optimise this alone like it does for 32 bits. */
	unsigned long n = hash;
	n <<= 18;
	hash -= n;
	n <<= 33;
	hash -= n;
	n <<= 3;
	hash += n;
	n <<= 3;
	hash -= n;
	n <<= 4;
	hash += n;
	n <<= 2;
	hash += n;
#else
	/* On some cpus multiply is faster, on others gcc will do shifts */
	hash *= GOLDEN_RATIO_PRIME;
#endif

	/* High bits are more random, so use them. */
	return hash >> (BITS_PER_LONG - bits);
}

#if BITS_PER_LONG == 32
#define HASH_RUSTY_SHIFT(X) (((X) << 13) | ((X) >> 19))
#else
#define HASH_RUSTY_SHIFT(X) ((X) << (BITS_PER_LONG - 32));
#endif

static u32 hash_rusty(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = hash_seed;
	res ^= key->sip;
	res ^= HASH_RUSTY_SHIFT(key->dip);
	res ^= (key->sport + (key->dport << 16)) << (BITS_PER_LONG - 32)/2;
	res = kernel_hash_long(res, htable_bits);
);
	return res;
}

/*
 * hash_abcd() has been proposed by Don Cohen
 * http://lists.samba.org/pipermail/netfilter-devel/2002-July/005008.html
 */
 
u32 coefn[][4] = { { 0x47441DFB, 0x57655A7D, 0x1C7F1D2D, 0xDF91D738 },
		   { 0x898092BB, 0x1CE6487C, 0xF219B787, 0x677F6D40 },
		   { 0x0E3A9C80, 0xE1338378, 0xB4CC63E7, 0x5BC57C1F },
		   { 0x0000173C, 0xF94471C8, 0x1EBFC863, 0xE7FBAE97 },
		   { 0x0000006A, 0x03FD9B60, 0xFDCFFE95, 0xFE32659F },

};
#define COEFN	((sizeof(coefn)/sizeof(coefn[0]))-1)

u32 *coef = coefn[0];

u64 coefll[4] = { 0x4476081a7043a46fL, 0x45768b8a6e7eac19L, 
  		  0xebd556c1cf055952L, 0x72ed2da1bf010101L };

static u32 hash_abcd(struct ct_key *key)
{
	u32 res; 
PER_HASH_TIMER_1(
	res = coef[0] * key->sip
	+ coef[1] * key->dip
	+ coef[2] * key->sport
	+ coef[3] * key->dport
	+ key->proto;
);
	return res;
}

static u32 hash_abcd64(struct ct_key *key)
{
  u64 ls, ld, lp;
  u32 sum;

PER_HASH_TIMER_1(
  ls = (u64) key->sip * coefll[0];
  ld = (u64) key->dip * coefll[1];
  lp = (u64)((u32)(key->sport << 16) + key->dport) * coefll[2] + key->proto;
  sum =
    ((ls ^ (ls >> 32)) +
     (ld ^ (ld >> 32)) +
     (lp ^ (lp >> 32))) ^ coefll[3];
);
  return sum;
}

/*
 * hash_abcdef() has later been proposed by Don Cohen; here it is
 * from code by Joakim.
 */
static u32 hash_abcdef(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ( 0x47441DFB * key->sip
	+ 0x57655A7D * key->dip
	+ 0x1C7F1D2D * key->sport
	+ 0xDF91D737 * key->dport
	+ 0x2F6FE8DB * key->proto ) ^ 0x39E5C2D7;
);
	return res;
}

/*
 * hash_abcd_xor(), misnamed abcd_power(), is by Joakim Axelsson and
 * Martin Josefsson, from a series of list mails starting here:
 * http://lists.samba.org/pipermail/netfilter-devel/2002-July/005068.html
 * I took all these functions from their working copy. Thanks.
 */
static u32 hash_abcd_xor(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = (0x47441DFB * key->sip)
	^ (0x57655A7D * key->dip)
	^ (0x1C7F1D2D * key->sport)
	^ (0xDF91D738 * key->dport)
	^ key->proto;
);
	return res;
}

/*
 * hash_rt() was derived from the IP routing code by Joakim and Martin.
 */
static u32 hash_rt(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->dip & 0xF0F0F0F0) >> 4) | ((key->dip & 0x0F0F0F0F) << 4);
	res ^= key->sip ^ key->proto;
	res ^= key->dport ^ key->sport;
	res ^= (res >> 24);
	res ^= (res >> 8);
);
	return res;
}

/*
 * hash_rt_ab() is an obfuscation of hash_rt(). The motto of this
 * function is "the compiler will certainly undo any constant damage."
 * (sorry, could not resists :)
 */
static u32 hash_rt_ab(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->dip & 0xF0F0F0F0) >> 4) | ((key->dip & 0x0F0F0F0F) << 4);
	res ^= key->sip ^ key->proto;
	res ^= key->dport ^ key->sport;
	res ^= 0x47441DFB;
	res ^= (res >> 24);
	res ^= (res >> 8);
	res ^= 0x57655A7D;
);
	return res;
}

/*
 * hash_rt_ab2() is a variant of hash_rt_ab().
 * http://lists.samba.org/pipermail/netfilter-devel/2002-July/005075.html
 */
static u32 hash_rt_ab2(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->sip & 0xF0F0F0F0) >> 4) | ((key->sip & 0x0F0F0F0F) << 4);
	res *= 0x47441DFB;
	res ^= key->dip ^ key->proto;
	res ^= ((key->dport << 16) | key->sport);
	res ^= 0x57655A7D;
	res ^= (res >> 24);
	res ^= (res >> 8);
);
	return res;
}

/*
 * hash_rt_new() is another variant of hash_rt_ab().
 */
static u32 hash_rt_new(struct ct_key *key)
{
	u32 res;
PER_HASH_TIMER_1(
	res = ((key->sip & 0xF0F0F0F0) >> 4) | ((key->sip & 0x0F0F0F0F) << 4);
	res *= 0x47441DFB;
	res ^= key->dip ^ key->proto;
	res ^= (key->dport << 16) ^ key->sport;
	res *= 0x57655A7D;
	res ^= (res >> 24);
	res ^= (res >> 8);
);
	return res;
}

/*
 * hash_crc32() uses the zlib crc32() function over the keys.
 */

#include "crc32.c"

static u32 hash_crc32(struct ct_key *key)
{
	u32 crc;
	u32 v32;
	u16 v16;
PER_HASH_TIMER_1(
	crc = crc32(hash_seed, (const char *) &key->proto, sizeof(key->proto));
	v32 = ntohl(key->sip);
	crc = crc32(crc, (const char *) &v32, sizeof(v32));
	v32 = ntohl(key->dip);
	crc = crc32(crc, (const char *) &v32, sizeof(v32));
	v16 = ntohs(key->sport);
	crc = crc32(crc, (const char *) &v16, sizeof(v16));
	v16 = ntohs(key->dport);
	crc = crc32(crc, (const char *) &v16, sizeof(v16));
);
	return crc;
}

/*
 * hash_random() simulates the optimum, by randomly choosing buckets.
 */
static u32 hash_random(struct ct_key *key)
{
	u32 res;
	PER_HASH_TIMER_1(res = (u32) random(););
	return res;
}

/*
 * 32-bit variant of the hash algorithm by Ron Guilmette 
 * (comp.compilers, 13 Dec 1990 06:25:10 GMT)
 */

static u32 hash_guilmette(struct ct_key *key)
{
	const char *str = (const char *) key;
	int i, j = 0, len = sizeof(struct ct_key);
	u32 res[4] = { 0, 0, 0, 0};
	
PER_HASH_TIMER_1(
	for (i = 0; i < len; i++, j++)
		res[j%4] = (str[i] ^ res[j%4]) * hash_seed;

	res[0] ^= res[3]<<24 ^ res[2]<<16 ^ res[1]<<8;
);
	return res[0];
}

/*
 * Algorithm by Bob Jenkins (http://burtleburtle.net/bob/hash/index.html)
 */
#define mix(a,b,c) \
{ \
  a -= b; a -= c; a ^= (c>>13); \
  b -= c; b -= a; b ^= (a<<8); \
  c -= a; c -= b; c ^= (b>>13); \
  a -= b; a -= c; a ^= (c>>12);  \
  b -= c; b -= a; b ^= (a<<16); \
  c -= a; c -= b; c ^= (b>>5); \
  a -= b; a -= c; a ^= (c>>3);  \
  b -= c; b -= a; b ^= (a<<10); \
  c -= a; c -= b; c ^= (b>>15); \
}

static u32 hash_jenkins(struct ct_key *key)
{
   unsigned char *k = (unsigned char *) key;
   u32 length = sizeof(struct ct_key);
   u32 a,b,c,len;

   /* Set up the internal state */
   len = length;
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = 0;           /* the previous hash value */

PER_HASH_TIMER_1(
   /*---------------------------------------- handle most of the key */
   while (len >= 12)
   {
      a += (k[0] +((u32)k[1]<<8) +((u32)k[2]<<16) +((u32)k[3]<<24));
      b += (k[4] +((u32)k[5]<<8) +((u32)k[6]<<16) +((u32)k[7]<<24));
      c += (k[8] +((u32)k[9]<<8) +((u32)k[10]<<16)+((u32)k[11]<<24));
      mix(a,b,c);
      k += 12; len -= 12;
   }

   /*------------------------------------- handle the last 11 bytes */
   c += length;
   switch(len)              /* all the case statements fall through */
   {
   case 11: c+=((u32)k[10]<<24);
   case 10: c+=((u32)k[9]<<16);
   case 9 : c+=((u32)k[8]<<8);
      /* the first byte of c is reserved for the length */
   case 8 : b+=((u32)k[7]<<24);
   case 7 : b+=((u32)k[6]<<16);
   case 6 : b+=((u32)k[5]<<8);
   case 5 : b+=k[4];
   case 4 : a+=((u32)k[3]<<24);
   case 3 : a+=((u32)k[2]<<16);
   case 2 : a+=((u32)k[1]<<8);
   case 1 : a+=k[0];
     /* case 0: nothing left to add */
   }
   mix(a,b,c);
   /*-------------------------------------------- report the result */
);

   return c;
}

/* The Jenkins hash fitted to connection tracking */
static u32 hash_jenkins2(struct ct_key *key)
{
   u32 a,b,c,len;

PER_HASH_TIMER_1(
   /* Set up the internal state */
   len = 4; /* source IP, dest IP, source port + proto, dest port + proto*/
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = 0;           /* the previous hash value */

   a += key->sip;
   b += ((u32)key->sport<<16) + (u32)key->proto;
   c += key->dip;
   mix(a,b,c);
   c += len;
   b += (u32)key->dport;
   mix(a,b,c);
);
   return c;
}

/* The Jenkins hash fitted to connection tracking with one mix. */
static u32 hash_jenkins2b(struct ct_key *key)
{
   u32 a,b,c,len;

PER_HASH_TIMER_1(
   /* Set up the internal state */
   len = 3; /* source IP, dest IP, source port + dest port + proto */
   a = b = 0x9e3779b9;  /* the golden ratio; an arbitrary value */
   c = 0;           /* the previous hash value */

   a += key->sip;
   b += ((u32)(key->dport + key->proto) << 16) + (u32)(key->sport + key->proto);
   c += key->dip;
   mix(a,b,c);
);
   return c;
}

/*
 * The 32-bit variant of the Pearson hash.
 */

/* Random numbers generated by www.random.org */
unsigned char rand4x256[4][256] = {
{28,219,206,227,193,71,3,43,199,222,22,61,188,132,182,226,
125,121,151,221,242,196,166,27,74,159,195,175,237,113,239,99,
207,161,167,174,185,84,231,253,158,118,142,0,139,213,98,11,
184,180,205,107,155,83,29,9,137,64,131,6,120,19,12,50,
140,225,1,143,172,30,128,48,252,191,202,247,14,51,232,254,
192,93,60,208,215,40,217,101,123,104,41,243,66,110,212,25,
135,117,13,156,178,81,255,17,224,90,133,62,223,145,95,250,
36,35,4,46,194,77,235,7,228,59,57,198,220,76,218,236,
58,79,126,89,209,171,148,97,229,108,210,146,109,176,241,189,
122,234,251,197,186,85,190,136,10,33,55,32,177,230,170,70,
69,100,63,8,245,144,54,150,39,44,203,103,162,94,31,153,
2,112,34,157,187,47,240,75,38,154,164,201,68,238,200,214,
138,73,147,149,152,246,127,124,37,78,42,88,114,45,233,15,
249,163,91,20,115,106,56,87,102,5,24,26,173,96,67,216,
119,134,211,23,16,204,72,53,129,105,49,82,18,181,244,169,
183,165,52,248,130,111,160,168,92,80,116,86,179,65,21,141},
{198,225,115,16,129,116,208,78,119,97,144,87,161,218,1,229,
234,58,223,142,76,145,92,244,108,231,101,45,39,199,104,204,
138,114,18,75,253,83,200,245,152,171,91,132,52,189,2,186,
176,26,249,148,209,8,0,143,3,41,197,113,182,252,126,251,
162,62,82,169,71,146,124,112,147,15,139,226,80,24,42,100,
32,69,121,150,192,117,81,167,215,141,65,109,219,177,213,205,
90,180,233,235,248,135,85,28,164,103,14,21,154,168,216,246,
89,217,33,110,111,107,224,22,66,122,188,30,44,170,250,163,
158,196,131,149,184,133,88,211,207,151,37,206,94,241,240,61,
99,57,190,195,166,17,20,120,70,51,19,228,55,191,181,48,
49,54,175,236,35,134,105,67,79,64,123,153,160,178,214,102,
203,10,201,60,56,38,68,185,242,5,77,12,128,137,118,157,
43,254,59,11,221,159,220,72,165,227,125,53,130,98,179,46,
47,194,7,4,155,25,106,50,9,183,31,202,232,156,23,127,
239,27,63,36,212,34,187,174,73,93,193,255,247,74,173,140,
95,86,6,210,13,172,136,96,222,243,40,230,29,84,238,237},
{179,146,31,22,48,195,213,11,14,116,12,60,41,174,133,8,
34,145,120,191,50,89,44,7,19,126,121,201,148,227,68,160,
173,82,180,88,177,85,96,15,76,199,140,139,13,233,231,217,
38,252,143,0,122,124,128,65,222,125,127,105,35,92,150,107,
104,170,64,210,90,244,228,232,94,189,115,238,153,193,251,42,
101,162,83,138,111,187,255,183,49,218,166,67,221,161,52,181,
204,33,1,16,98,151,246,63,164,235,54,134,245,219,155,108,
188,73,247,95,229,58,158,169,102,239,250,236,86,2,10,91,
249,84,4,243,167,137,100,241,109,113,118,53,240,30,154,36,
175,194,209,176,27,224,26,171,225,28,214,75,156,163,135,226,
216,5,3,77,20,152,97,61,129,223,207,32,57,186,59,112,
254,103,71,23,185,72,165,45,39,230,46,87,80,51,24,74,
25,6,106,142,21,190,192,79,212,159,17,215,147,18,253,168,
47,242,56,66,55,141,157,9,123,234,202,40,78,69,184,237,
70,114,200,206,178,117,136,208,81,196,110,131,203,43,29,149,
119,132,144,211,220,182,172,130,99,248,37,93,198,205,62,197},
{171,56,191,163,255,46,251,71,92,100,30,149,127,208,250,217,
233,132,38,141,62,95,131,13,59,181,196,221,175,16,164,236,
108,105,235,178,125,135,201,14,193,122,151,237,238,198,176,104,
123,130,36,57,83,19,93,133,229,45,134,3,101,114,37,143,
117,67,183,157,243,102,34,254,53,244,210,32,253,107,48,86,
220,190,15,109,137,139,22,96,35,31,120,89,156,7,24,231,
103,61,66,33,111,64,79,227,10,84,160,52,186,129,159,68,
29,204,225,63,189,147,40,97,184,28,207,168,187,70,88,5,
182,228,161,240,121,60,4,166,18,77,115,197,112,1,21,239,
218,185,249,41,90,17,199,44,232,194,222,173,87,69,49,82,
219,165,150,118,65,195,211,75,26,145,214,25,98,78,142,136,
42,58,200,213,252,230,11,76,177,212,106,126,81,248,215,85,
158,27,116,55,73,0,174,74,47,119,202,39,6,169,206,148,
167,242,155,12,226,154,2,50,192,138,172,224,94,188,128,170,
51,113,20,43,99,223,124,245,216,203,110,54,205,241,152,8,
144,80,146,91,153,246,140,179,180,162,247,23,209,72,9,234}
};

static u32 hash_pearson(struct ct_key *key)
{
	const char *str = (const char *) key;
	int i, j, len = sizeof(struct ct_key);
	u32 res[4] = { 0, 0, 0, 0};
	
PER_HASH_TIMER_1(
	for (i = 0; i < len; i++)
		for (j = 0; j < 4; j++)
			res[j] = rand4x256[j][res[j] ^ str[i]];

	res[0] ^= res[3]<<24 ^ res[2]<<16 ^ res[1]<<8; 
);
	return res[0];
}

struct hash_def {
	char *name;
	u32 (*hash_f)(struct ct_key *);
	u32 seed;
};

#define NR_HASHES (sizeof(hashes)/sizeof(hashes[0]))

static struct hash_def hashes[] = {
	{ "original",	hash_conntrack, 0 },
	{ "rusty",	hash_rusty,	0 },
	{ "abcd",	hash_abcd,	0 },
	{ "abcdef",	hash_abcdef,	0 },
	{ "abcd_xor",	hash_abcd_xor,	0 },
	{ "rt",		hash_rt,	0 },
	{ "rt_ab",	hash_rt_ab,	0 },
	{ "rt_ab2",	hash_rt_ab2,	0 },
	{ "rt_new",	hash_rt_new,	0 },
	{ "crc32",	hash_crc32,	0 },
	{ "random",	hash_random,	0 },
	{ "guilmette",	hash_guilmette,	47 },
	{ "jenkins",	hash_jenkins, 0 },
	{ "jenkins2",	hash_jenkins2, 0 },
	{ "jenkins2b",	hash_jenkins2b, 0 },
	{ "pearson",	hash_pearson, 0 },
	{ "abcd64",	hash_abcd64, 0 },
};

static u32 *occupation;
static struct hash_def *test_hash = hashes;

static void do_init(void)
{
	u32 idx;
	occupation = malloc(sizeof(*occupation) * htable_size);
	for (idx=0; idx<htable_size; idx++)
		occupation[idx] = 0;
}

static int display_target = -1;

static int do_count(struct ct_key *key)
{
	u32 idx;
	
	PER_HASH_TIMER_2(idx = test_hash->hash_f(key) % htable_size;);
	occupation[idx]++;
	return ((int)idx) == display_target;
}

static u32 parse_ip(char *t)
{
	char *p = strchr(t, '=');
	if (!p++) return 0;
	return inet_addr(p);
}

static u16 parse_port(char *t)
{
	char *p = strchr(t, '=');
	if (!p++) return 0;
	return htons((u16) strtoul(p, 0, 10));
}

static u32 ignore_tcp_timeout_bound = 0;

static int ignore_by_timeout(struct ct_key *key, char *tok)
{
	return	key->proto == 6
		&& strtoul(tok, 0, 10) < ignore_tcp_timeout_bound;
}

static int unidir;

static int process(char *p)
{
	struct ct_key key[1];
	char *tok;
	int res;

	strtok(p, " \t\n");					/* prot text */
	tok = strtok(0, " \t\n");
	key->proto = (u16) strtoul(tok, 0, 10);			/* prot num */
	tok = strtok(0, " \t\n");				/* timeout */
	if (ignore_by_timeout(key, tok)) return 0;
	switch (key->proto) {
		case 6:  strtok(0, " \t\n");			/* tcp state */
		case 17: break;
		default: return 0;
	}
	/* forward */
	tok = strtok(0, " \t\n"); key->sip = parse_ip(tok);	/* src ip */
	tok = strtok(0, " \t\n"); key->dip = parse_ip(tok);	/* dst ip */
	tok = strtok(0, " \t\n"); key->sport = parse_port(tok);	/* src port */
	tok = strtok(0, " \t\n"); key->dport = parse_port(tok);	/* dst port */
	res = do_count(key);
	if (unidir) return res;
	/* reverse */
	tok = strtok(0, " \t\n");
	if (*tok == '[') tok = strtok(0, " \t\n");
	key->sip = parse_ip(tok);				/* src ip */
	tok = strtok(0, " \t\n"); key->dip = parse_ip(tok);	/* dst ip */
	tok = strtok(0, " \t\n"); key->sport = parse_port(tok);	/* src port */
	tok = strtok(0, " \t\n"); key->dport = parse_port(tok);	/* dst port */
	return res | do_count(key);
}

static int occupation_target = -1;

/*
 *	              n
 *	P(0) = (1 - p)
 *
 *	          k   n - i     p             n
 *	P(k+1) = prod(----- * -----) * (1 - p)
 *	         i=0  i + 1   1 - p
 */

static long double probab(u32 n, u32 b, u32 k)
{
	long double res = powl(1.0L - 1.0L/b, (long double)n);
	int i;
	
	for (i = 0; i < k; i++)
		res *= ((long double)n - (long double)i)/((long double)b*(1.0L+(long double)i)*(1.0L-1.0L/b));
		
	return res;
}

/*
 * Check that every input bit changes every output bit half the time
 * from lookup2.c by Bob Jenkins
 */

#define HASHSTATE 1
#define HASHLEN   1
#define MAXPAIR 80
#define MAXLEN 14+2		/* padding */
static void jenkins_test(void) {
	unsigned char qa[MAXLEN+1], qb[MAXLEN+2], *a = &qa[0], *b = &qb[1];
	u32 c[HASHSTATE], d[HASHSTATE], i, j=0, k, l, m, z;
	u32 e[HASHSTATE],f[HASHSTATE],g[HASHSTATE],h[HASHSTATE];
	u32 x[HASHSTATE],y[HASHSTATE];
	u32 hlen;

	printf("Tested hash: %s, %i\n", test_hash->name, sizeof(struct ct_key));
	printf("No more than %d trials should ever be needed \n",MAXPAIR/2);
	for (hlen = 0; hlen < MAXLEN; ++hlen) {
		z=0;
    		for (i=0; i<hlen; ++i)  {
    			if (i == 6 || i == 7)
    				continue;	/* padding */
    			/*----------------------- for each input byte, */
      			for (j=0; j<8; ++j) {  
      				/*------------------------ for each input bit, */
				for (m=1; m<8; ++m) {
					/*------------ for serveral possible initvals, */
	  				for (l=0; l<HASHSTATE; ++l) 
	  					e[l]=f[l]=g[l]=h[l]=x[l]=y[l]=~((u32)0);

	      	  			/*---- check that every output bit is affected by that input bit */
		  			for (k=0; k<MAXPAIR; k+=2) { 
	    					u32 finished=1;
	    					/* keys have one bit different */
	    					for (l=0; l<hlen+1; ++l) { 
	    						a[l] = b[l] = (unsigned char)0;
	    					}
	    					/* have a and b be two keys differing in only one bit */
	    					a[i] ^= (k<<j);
	    					a[i] ^= (k>>(8-j));
	    					c[0] = test_hash->hash_f((struct ct_key *)a);
	    					b[i] ^= ((k+1)<<j);
	    					b[i] ^= ((k+1)>>(8-j));
	     					d[0] = test_hash->hash_f((struct ct_key *)b);
	    					/* check every bit is 1, 0, set, and not set at least once */
	    					for (l=0; l<HASHSTATE; ++l) {
	      						e[l] &= (c[l]^d[l]);
	      				 		f[l] &= ~(c[l]^d[l]);
	      						g[l] &= c[l];
	      						h[l] &= ~c[l];
	      						x[l] &= d[l];
	      						y[l] &= ~d[l];
	      						if (e[l]|f[l]|g[l]|h[l]|x[l]|y[l]) 
	      							finished=0;
	    					}
	    					if (finished)
	    						break;
	  				}
	  				if (k>z) 
	  					z=k;
	  				if (k==MAXPAIR) {

	     					printf("Some bit didn't change: ");
	     					printf("%.8lx %.8lx %.8lx %.8lx %.8lx %.8lx  ",
	            					e[0],f[0],g[0],h[0],x[0],y[0]);
	     					printf("i %ld j %ld m %ld len %ld\n",i,j,m,hlen);
	  				}
	  				if (z==MAXPAIR) 
	  					goto done;
				}
			}
      		}
    	}
   	done:
    	if (z < MAXPAIR) {
      		printf("Mix success  %2ld bytes  %2ld initvals  ",i,m);
      		printf("required  %ld  trials\n",z/2);
    	}
  	printf("\n");
}

static void show_occupation(void)
{
	u32 idx, maxlen, use, *agg;
	u64 partial;
	int n, mult;

	for (idx=0, maxlen=0, use=0; idx<htable_size; idx++) {
		if (occupation[idx] > 0) use++;
		if (occupation[idx] > maxlen) {
			mult = 1;
			maxlen = occupation[idx];
		} else if (occupation[idx] == maxlen)
			mult++;
	}

	printf("total buckets:       %lu\n", htable_size);
	printf("total tuples:        %lu (%.02f%%)\n", total_hash_count, (100.0*total_hash_count)/htable_size);
	printf("used buckets:        %lu (%.02f%%)\n", use, (100.0*use)/htable_size);
	printf("maximum list length: %lu\n", maxlen);
	printf("average list length: %.02f\n", (1.0 * total_hash_count) / htable_size);
	printf("probability of one longest bucket length: %.6LE (%u)\n",
		probab(total_hash_count, htable_size, maxlen), mult);
#ifdef PER_HASH_OUTPUT
	printf("hash cost per tuple: %Lu cyc, %.0f ns, %.0f Mhz on %.0f Mhz CPU (%s)\n",
		total_hash_time / total_hash_count,
		(ns_per_clock * total_hash_time) / total_hash_count,
		(1e3 * total_hash_count) / (ns_per_clock * total_hash_time),
		1e3 / ns_per_clock,
		PER_HASH_OUTPUT_TYPE);
#endif
	agg = malloc(sizeof(*agg) * (maxlen+1));
	for (idx=0; idx<=maxlen; idx++)
		agg[idx] = 0;
	for (idx=0; idx<htable_size; idx++) {
		agg[occupation[idx]]++;
		if (((int)occupation[idx]) == occupation_target) {
			printf("occupation target %d at %lu\n",
				occupation_target, idx);
			occupation_target = -1;
		}
	}
	printf("Count\tLength\tCT\n");
	for (partial=0ULL, idx=maxlen; idx>0; idx--) {
		double d;
		if (agg[idx] == 0) continue;
		partial += idx * agg[idx];
		d = (100.0 * partial) / (1.0 * total_hash_count);
		printf("%lu\t%lu\t%.02f%%\n", agg[idx], idx, d);
	}
	printf("%lu\t0\t100.00%%\n", agg[idx]);
}

static struct hash_def *find_hash(char *name)
{
	u32 idx;
	char *num = 0;
	char *p = name + strlen(name) - 1;

	if ((p-1) >= name && p[-1] == '+' && (*p == 'r' || *p == 'R')) {
		p[-1] = '\0';
		srandom(time(0));
		hash_seed = random();
	} else {
		while (p > name && *p >= '0' && *p <= '9')
			num = p--;
		if (num && p >= name) {
			if (*p == '+') {
				hash_seed = strtoul(num, 0, 10);
				*p = '\0';
			} else if (*p == '-' && strncmp(name, "abcd-", 5) == 0) {
				idx = strtoul(num, 0, 10);
				if (idx > COEFN) {
					fprintf(stderr, "no hash function named '%s'\n", name);
					return 0;
				}
				*p = '\0';
				coef = coefn[idx];
			}
		}
	}
	for (idx=0; idx<NR_HASHES; idx++) {
		if (0 == strcmp(name, hashes[idx].name)) {
			if (!hash_seed)
				hash_seed = hashes[idx].seed;
			return hashes+idx;
		}
	}
	fprintf(stderr, "no hash function named '%s'\n", name);
	return 0;
}

static void usage(void)
{
	fprintf(stderr, "Usage: cttest [-f filename] [-s hashsize] [-h hashfunc-type]\n");
	fprintf(stderr, "              [-i dtarget] [-t otarget] [-b tcp-timeout]\n");
	fprintf(stderr, "              [-c hash-seed] [-C] [-P] [-u]\n");
	exit(1);
}

#include "prime.c"

static u32 size2bits(u32 size)
{
	u32 bits;
	for (size--,bits=0; size; bits++, size>>=1) /**/ ;
	return bits;
}

int main(int argc, char **argv)
{
	int c;
	char buf[512], copy[512];
	char *filename = "/proc/net/ip_conntrack";
	FILE *fp;
	int select_prime = 0, bit_test = 0;

	while (EOF!=(c=getopt(argc, argv, "f:s:h:i:t:b:c:CPuB"))) switch(c) {
		case 'f': filename = optarg; break;
		case 'u': unidir = 1; break;
		case 's': htable_size = strtoul(optarg, 0, 10); break;
		case 'P': select_prime = 1; break;
		case 'B': bit_test = 1; break;
		case 'h': test_hash = find_hash(optarg); break;
		case 'i': display_target = strtol(optarg, 0, 10); break;
		case 't': occupation_target = strtol(optarg, 0, 10); break;
		case 'b': ignore_tcp_timeout_bound = strtoul(optarg, 0, 10); break;
		case 'c': hash_seed = strtoul(optarg, 0, 10); break;
		case 'C': srandom(time(0)); hash_seed = random(); break;
		default: usage();
	}
	if (!test_hash) return 1;
	if (bit_test) {
		jenkins_test();
		exit(0);
	}
	fp = (0 == strcmp(filename, "-")) ? stdin : fopen(filename, "r");
	if (!fp) {
		fprintf(stderr, "%s: %s\n", filename, strerror(errno));
		exit(1);
	}
#ifdef PER_HASH_OUTPUT
	rdtsc_cal(100000000);
#endif
	if (select_prime) htable_size = make_prime(htable_size);
	htable_bits = size2bits(htable_size);
	printf("filename: %s\n", filename);
	printf("hash function: %s\n", test_hash->name);
	if (ignore_tcp_timeout_bound)
		printf("ignore TCP conntracks with timeout less than %lu\n",
			ignore_tcp_timeout_bound);
	if (hash_seed)
		printf("hash seed 0x%0lx\n", hash_seed);
	do_init();
	while (fgets(buf, sizeof(buf), fp)) {
		if (display_target != -1)
			strcpy(copy, buf);
		if (process(buf))
			printf("[%d]\t%s", display_target, copy);
	}
	if (fp != stdin) fclose(fp);
	show_occupation();
	return 0;
}
